// SPDX-License-Identifier: BSD-3-Clause

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/epoll.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/sendfile.h>
#include <sys/eventfd.h>
#include <libaio.h>
#include <errno.h>

#include "aws.h"
#include "utils/util.h"
#include "utils/debug.h"
#include "utils/sock_util.h"
#include "utils/w_epoll.h"

/* server socket file descriptor */
static int listenfd;

/* epoll file descriptor */
static int epollfd;

static io_context_t ctx;

static int aws_on_path_cb(http_parser *p, const char *buf, size_t len)
{
	struct connection *conn = (struct connection *)p->data;

	memcpy(conn->request_path, buf, len);
	conn->request_path[len] = '\0';
	conn->have_path = 1;

	return 0;
}

static void connection_prepare_send_reply_header(struct connection *conn)
{
	/* TODO: Prepare the connection buffer to send the reply header. */

	snprintf(conn->send_buffer, sizeof(conn->send_buffer),
			"HTTP/1.0 200 OK\r\n"
			"Content-Length: %zu\r\n"
			"Connection: close\r\n"
			"\r\n",
			(size_t)conn->file_size);

	conn->send_len = strlen(conn->send_buffer);
	conn->send_pos = 0;
}

static void connection_prepare_send_404(struct connection *conn)
{
	/* TODO: Prepare the connection buffer to send the 404 header. */

	snprintf(conn->send_buffer, sizeof(conn->send_buffer),
			"HTTP/1.1 404 Not Found\r\n"
			"Content-Length: 0\r\n"
			"Connection: close\r\n"
			"\r\n");

	conn->send_len = strlen(conn->send_buffer);
	conn->send_pos = 0;
}

static enum resource_type connection_get_resource_type(struct connection *conn)
{
	/* TODO: Get resource type depending on request path/filename. Filename should
	 * point to the static or dynamic folder.
	 */

	if (strstr(conn->request_path, "static") != NULL)
		return RESOURCE_TYPE_STATIC; /* static */

	if (strstr(conn->request_path, "dynamic") != NULL)
		return RESOURCE_TYPE_DYNAMIC; /* dinamic */

	return RESOURCE_TYPE_NONE; /* error */
}


struct connection *connection_create(int sockfd)
{
	/* TODO: Initialize connection structure on given socket. */

	/* Allocate memory for the connection structure. */
	struct connection *conn = malloc(sizeof(struct connection));

	/* Initialize connection fields. */
	conn->sockfd = sockfd;
	conn->have_path = 0;
	conn->recv_len = 0;
	conn->send_len = 0;
	conn->send_pos = 0;
	conn->file_pos = 0;
	conn->async_read_len = 0;
	conn->file_size = 0;
	conn->fd = -1;
	conn->eventfd = -1;
	conn->state = STATE_NO_STATE;
	memset(conn->request_path, 0, sizeof(conn->request_path));
	memset(conn->recv_buffer, 0, sizeof(conn->recv_buffer));
	memset(conn->send_buffer, 0, sizeof(conn->send_buffer));

	/* Prepare the AIO structure */
	conn->ctx = ctx;
	conn->iocb.data = conn;
	conn->piocb[0] = &conn->iocb;

	conn->eventfd = eventfd(0, EFD_NONBLOCK | EFD_CLOEXEC);
	w_epoll_add_ptr_in(epollfd, conn->eventfd, conn);

	/* Initialize HTTP_REQUEST parser. */
	http_parser_init(&conn->request_parser, HTTP_REQUEST);
	conn->request_parser.data = conn;

	return conn;
}

void connection_start_async_io(struct connection *conn)
{
	if (conn->file_pos >= conn->file_size)
		return;

	size_t read_size = BUFSIZ;

	if (conn->file_size - conn->file_pos < BUFSIZ)
		read_size = conn->file_size - conn->file_pos;

	io_prep_pread(&conn->iocb, conn->fd, conn->send_buffer, read_size, conn->file_pos);

	/* AIO notify us when done */
	io_set_eventfd(&conn->iocb, conn->eventfd);

	int ret = io_submit(conn->ctx, 1, conn->piocb);

	if (ret < 0) {
		connection_remove(conn);
		return;
	}

	conn->state = STATE_ASYNC_ONGOING;
}

void connection_remove(struct connection *conn)
{
	/* remove connection from epoll */
	if (conn->sockfd != -1) {
		w_epoll_remove_ptr(epollfd, conn->sockfd, conn);
		close(conn->sockfd);
		conn->sockfd = -1;
	}

	/* close fd */
	if (conn->fd != -1) {
		close(conn->fd);
		conn->fd = -1;
	}

	/* close eventfd */
	if (conn->eventfd != -1) {
		w_epoll_remove_ptr(epollfd, conn->eventfd, conn);
		close(conn->eventfd);
		conn->eventfd = -1;
	}

	conn->state = STATE_CONNECTION_CLOSED;
	free(conn);
}

void handle_new_connection(void)
{
	/* TODO: Handle a new connection request on the server socket. */
	socklen_t addrlen = sizeof(struct sockaddr_in);
	struct sockaddr_in addr;

	/* Accept new connection. */
	int connectfd = accept(listenfd, (SSA *)&addr, &addrlen);

	/* Set socket to be non-blocking. */
	fcntl(connectfd, F_SETFL, O_NONBLOCK);

	/* Instantiate new connection handler. */
	struct connection *conn = connection_create(connectfd);

	/* Add socket to epoll. */
	w_epoll_add_ptr_in(epollfd, connectfd, conn);

	/* Initialize HTTP_REQUEST parser. */
	http_parser_init(&conn->request_parser, HTTP_REQUEST);
}

void receive_data(struct connection *conn)
{
	/* TODO: Receive message on socket.
	 * Store message in recv_buffer in struct connection.
	 */

	ssize_t len = 0;

	/* reading from the client */
	while ((len = recv(conn->sockfd, conn->recv_buffer + conn->recv_len, BUFSIZ, 0)) > 0)
		conn->recv_len += len;

	conn->recv_buffer[conn->recv_len] = '\0';

	// dlog(LOG_DEBUG, "Received data: %s\n", conn->recv_buffer);

	if (len < 0) {
		if (errno == EAGAIN || errno == EWOULDBLOCK)
			return;

		ERR("Error receiving data\n");
		connection_remove(conn);
	}

	conn->state = STATE_REQUEST_RECEIVED;
}

int connection_open_file(struct connection *conn)
{
	/* TODO: Open file and update connection fields. */

	const char *relative_path = conn->request_path;
	char path[2 * BUFSIZ];

	snprintf(path, sizeof(path), ".%s", relative_path);

	// fprintf(stderr, "Opening file: %s\n", path);

	/* opening the file */
	conn->fd = open(path, O_RDONLY);
	if (conn->fd < 0)
		return -1;

	/* getting the file size */
	struct stat st;

	fstat(conn->fd, &st);
	conn->file_size = st.st_size;

	return 0;
}

void connection_complete_async_io(struct connection *conn)
{
	uint64_t event_count;
	ssize_t s = read(conn->eventfd, &event_count, sizeof(event_count));

	if (s != sizeof(event_count)) {
		connection_remove(conn);
		return;
	}

	/* retrieve the events */
	struct io_event events[1];
	int ret = io_getevents(conn->ctx, 1, 1, events, NULL);

	if (ret < 0) {
		connection_remove(conn);
		return;
	}

	ssize_t bytes_read = events[0].res;

	/* read `bytes_read` bytes into conn->send_buffer. */
	conn->send_len = bytes_read;
	conn->send_pos = 0;
	conn->file_pos += bytes_read;

	/* now we have a chunk ready to send. */
	conn->state = STATE_SENDING_DATA;
}

int parse_header(struct connection *conn)
{
	/* TODO: Parse the HTTP header and extract the file path. */
	/* Use mostly null settings except for on_path callback. */

	http_parser_settings settings_on_path = {
		.on_message_begin = 0,
		.on_header_field = 0,
		.on_header_value = 0,
		.on_path = aws_on_path_cb,
		.on_url = 0,
		.on_fragment = 0,
		.on_query_string = 0,
		.on_body = 0,
		.on_headers_complete = 0,
		.on_message_complete = 0
	};

	conn->request_parser.data = conn;
	http_parser_execute(&conn->request_parser, &settings_on_path, conn->recv_buffer, conn->recv_len);

	return 0;
}

enum connection_state connection_send_static(struct connection *conn)
{
	off_t offset = conn->file_pos;

	/* loop which sends data */
	while (conn->send_len > 0) {
		ssize_t sent = sendfile(conn->sockfd, conn->fd, &offset, conn->send_len);

		if (sent < 0) {
			if (errno == EAGAIN || errno == EWOULDBLOCK)
				return STATE_SENDING_DATA;

			connection_remove(conn);
			return STATE_CONNECTION_CLOSED;

		} else {
			/* updating the pos and length of the file */
			conn->file_pos += sent;
			conn->send_len -= sent;

			/* file is sent */
			if (conn->file_pos >= conn->file_size) {
				close(conn->fd);
				conn->fd = -1;
				return STATE_DATA_SENT;
			}
		}
	}

	return STATE_DATA_SENT;
}

int connection_send_data(struct connection *conn)
{
	/* May be used as a helper function. */
	/* TODO: Send as much data as possible from the connection send buffer.
	 * Returns the number of bytes sent or -1 if an error occurred
	 */

	if (conn->res_type == RESOURCE_TYPE_STATIC) {
		enum connection_state state = connection_send_static(conn);

		if (state == STATE_DATA_SENT)
			return 0; /* succes */

		else if (state == STATE_SENDING_DATA)
			return 1; /* partially */

		return -1; /* error */

	} else if (conn->res_type == RESOURCE_TYPE_DYNAMIC) {
		int state = connection_send_dynamic(conn);

		if (state == STATE_DATA_SENT)
			return 0; /* succes */

		else if (state == STATE_SENDING_DATA || state == STATE_ASYNC_ONGOING)
			return 1; /* partially */

		return -1; /* error */

	} else {
		return -1;
	}
}


int connection_send_dynamic(struct connection *conn)
{
	while (conn->send_pos < conn->send_len) {
		ssize_t sent = send(conn->sockfd,
							conn->send_buffer + conn->send_pos,
							conn->send_len - conn->send_pos,
							0);

		if (sent < 0) {
			if (errno == EAGAIN || errno == EWOULDBLOCK)
				return STATE_SENDING_DATA; /* partial */

			connection_remove(conn);
			return STATE_CONNECTION_CLOSED;
		}

		conn->send_pos += sent;
	}

	/* finished this chunk, check if there's more to read */
	if (conn->file_pos < conn->file_size) {
		/* reading the next chunk */
		connection_start_async_io(conn);
		return STATE_ASYNC_ONGOING;
	}

	/* No more data to read => we are done sending this file */
	return STATE_DATA_SENT;
}

void handle_input(struct connection *conn)
{
    /* receive data from the client */
	receive_data(conn);

    /* already have a path */
	if (conn->have_path)
		return;

    /* parse */
	parse_header(conn);

	/* no path */
	if (!conn->have_path) {
		connection_prepare_send_404(conn);
		w_epoll_update_ptr_out(epollfd, conn->sockfd, conn);
		conn->state = STATE_SENDING_404;
		return;
	}

	/* resource type and open file. */
	conn->res_type = connection_get_resource_type(conn);
	if (connection_open_file(conn) < 0) {
		connection_prepare_send_404(conn);
		w_epoll_update_ptr_out(epollfd, conn->sockfd, conn);
		conn->state = STATE_SENDING_404;
		return;
	}

	if (conn->res_type == RESOURCE_TYPE_STATIC) {
		/* The resource is static. */
		connection_prepare_send_reply_header(conn);
		w_epoll_update_ptr_out(epollfd, conn->sockfd, conn);
		conn->state = STATE_SENDING_HEADER;

	} else if (conn->res_type == RESOURCE_TYPE_DYNAMIC) {
		/* The resource is dynamic. */
		connection_prepare_send_reply_header(conn);
		w_epoll_update_ptr_out(epollfd, conn->sockfd, conn);
		conn->state = STATE_SENDING_HEADER;

	} else {
		/* The resource does not exist. */
		connection_prepare_send_404(conn);
		w_epoll_update_ptr_out(epollfd, conn->sockfd, conn);
		conn->state = STATE_SENDING_404;
	}
}

void handle_output(struct connection *conn)
{
	switch (conn->state) {
	case STATE_SENDING_HEADER: {
		/* Send header first */
		ssize_t len = send(conn->sockfd,
							conn->send_buffer + conn->send_pos,
							conn->send_len - conn->send_pos,
							0);

		if (len < 0) {
			if (errno == EAGAIN || errno == EWOULDBLOCK)
				return; /* partial */

			ERR("Error sending data\n");
			connection_remove(conn);
			return;
		}

		conn->send_pos += len;
		if (conn->send_pos < conn->send_len)
			return; /* partial */

		/* done sending header => resource type */
		if (conn->res_type == RESOURCE_TYPE_STATIC) {
			conn->send_len = conn->file_size;
			conn->send_pos = 0;
			conn->state = STATE_SENDING_DATA;
			w_epoll_update_ptr_out(epollfd, conn->sockfd, conn);

		} else if (conn->res_type == RESOURCE_TYPE_DYNAMIC) {
			connection_start_async_io(conn);
			conn->state = STATE_SENDING_DATA;
			w_epoll_update_ptr_out(epollfd, conn->sockfd, conn);

		} else {
			connection_remove(conn);
		}

		break;
	}

	case STATE_SENDING_DATA: {
		int ret = connection_send_data(conn);

		if (ret < 0) {
			connection_remove(conn); /* error */

		} else if (ret == 0) {
			/* done */
			connection_remove(conn);

		} else {
			if (conn->state == STATE_ASYNC_ONGOING)
				return;

			w_epoll_update_ptr_out(epollfd, conn->sockfd, conn);
		}

		break;
	}

	case STATE_SENDING_404: {
		/* Send 404 buffer */
		ssize_t len = send(conn->sockfd,
							conn->send_buffer + conn->send_pos,
							conn->send_len - conn->send_pos,
							0);

		if (len < 0) {
			if (errno == EAGAIN || errno == EWOULDBLOCK)
				return; /* partial */

			ERR("Error sending data\n");
			connection_remove(conn);
			return; /* error */
		}

		conn->send_pos += len;
		if (conn->send_pos < conn->send_len)
			return; /* partial */

		/* 404 done => close */
		connection_remove(conn);
		break;
	}

	default:
		connection_remove(conn);
		break;
	}
}

void handle_client(uint32_t event, struct connection *conn)
{
	/* async in progress => complete it */
	if (conn->state == STATE_ASYNC_ONGOING && (event & EPOLLIN)) {
		connection_complete_async_io(conn);

		if (conn->sockfd == -1)
			return; // connection was removed
	}

	if (event & EPOLLIN)
		handle_input(conn);

	if (event & EPOLLOUT)
		handle_output(conn);
}

int main(void)
{
	int rc;

	/* TODO: Initialize asynchronous operations. */
	rc = io_setup(128, &ctx);
	if (rc < 0) {
		ERR("io\n");
		return 1;
	}

	/* TODO: Initialize multiplexing. */
	epollfd = w_epoll_create();

	/* TODO: Create server socket. */
	listenfd = tcp_create_listener(AWS_LISTEN_PORT, DEFAULT_LISTEN_BACKLOG);

	int flags = fcntl(listenfd, F_GETFL, 0);

	flags |= O_NONBLOCK;
	fcntl(listenfd, F_SETFL, flags);

	/* TODO: Add server socket to epoll object*/
	w_epoll_add_fd_in(epollfd, listenfd);

	/* Uncomment the following line for debugging. */
	// dlog(LOG_INFO, "Server waiting for connections on port %d\n", AWS_LISTEN_PORT);

	/* server main loop */
	while (1) {
		struct epoll_event rev;

		/* TODO: Wait for events. */
		rc = w_epoll_wait_infinite(epollfd, &rev);

		/* TODO: Switch event types; consider
		 *   - new connection requests (on server socket)
		 *   - socket communication (on connection sockets)
		 */

		if (rev.data.fd == listenfd) {
			handle_new_connection();

		} else {
			struct connection *conn = rev.data.ptr;

			handle_client(rev.events, conn);
		}
	}

	close(epollfd);
	close(listenfd);
	io_destroy(ctx);

	return 0;
}
